-- Lua optimizations
local ceil = ceil
local date = date
local format = format
local gsub = gsub
local strfind = strfind
local strlen = strlen
local strsplit = strsplit
local strsub = strsub
local strupper = strupper
local time = time
local tonumber = tonumber
local GetRealmName = GetRealmName
local GetNumSavedInstances = GetNumSavedInstances
local GetSavedInstanceInfo = GetSavedInstanceInfo
local SecondsToTime = SecondsToTime
local RequestRaidInfo = RequestRaidInfo
local UnitClass = UnitClass
local UnitName = UnitName

-- Use Blizzard's strings to save localization efforts
local NO_RAID_INSTANCES_SAVED = NO_RAID_INSTANCES_SAVED -- "You are not saved to any instances"

local Indicators = {
	ICON_STAR = ICON_LIST[1].."16:16:0:0|t",
	ICON_CIRCLE = ICON_LIST[2].."16:16:0:0|t",
	ICON_DIAMOND = ICON_LIST[3].."16:16:0:0|t",
	ICON_TRIANGLE = ICON_LIST[4].."16:16:0:0|t",
	ICON_MOON = ICON_LIST[5].."16:16:0:0|t",
	ICON_SQUARE = ICON_LIST[6].."16:16:0:0|t",
	ICON_CROSS = ICON_LIST[7].."16:16:0:0|t",
	ICON_SKULL = ICON_LIST[8].."16:16:0:0|t",
	BLANK = "None",
	TEXT = "Text",
}



local SavedInstances = LibStub("AceAddon-3.0"):NewAddon("SavedInstances", "AceEvent-3.0")
_G.SavedInstances = SavedInstances
local Config = LibStub("AceConfig-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("SavedInstances")

local LDB = LibStub("LibDataBroker-1.1", true)
local LDBIcon = LDB and LibStub("LibDBIcon-1.0", true)
local launcher

local QTip = LibStub("LibQTip-1.0")
local tooltip

local selectedInstance -- for re-ordering the instance list

local db
local defaultDB = {
	DBVersion = 4,
	Toons = { }, 	-- table key: "Toon - Realm"; value:
				-- Class: string
				-- Hide: boolean
				-- NumSaved: integer
	Indicators = {
		NormalIndicator = "ICON_STAR", -- indicator: ICON_*, BLANK, TEXT
		NormalText = "N",
		NormalColor = { 0, 1, 0, 1, }, -- green
		HeroicIndicator = "ICON_SKULL", -- indicator
		HeroicText = "H",
		HeroicColor = { 1, 0, 0, 1, }, -- red
		UnsavedIndicator = "BLANK", -- indicator
		UnsavedText = "",
		UnsavedColor = { 1, 1, 1, 1, }, -- white
	},
	Tooltip = {
		FreshToons = false,
		AllRaids = true,
		Details = false,
		NewInstanceShow = false,
	},
	Instances = { }, 	-- table key: "Instance name"; value:
					-- Show: boolean
					-- Order: integer
					-- table key: "Toon - Realm"; value:
						-- table key: "Difficulty"; value:
							-- ID: integer
							-- Expires: integer
	MinimapIcon = { },
}

-- general helper functions below

local function ColorCodeOpen(color)
	local r = format("%x", math.floor(color[1] * 255))
	if strlen(r) == 1 then
		r = "0"..r
	end
	local g = format("%x", math.floor(color[2] * 255))
	if strlen(g) == 1 then
		g = "0"..g
	end
	local b = format("%x", math.floor(color[3] * 255))
	if strlen(b) == 1 then
		b = "0"..b
	end
	local a = format("%x", math.floor(color[4] * 255))
	if strlen(a) == 1 then
		a = "0"..a
	end
	return "|c"..a..r..g..b
end

local function ClassColorise(class, targetstring)
	local class = strupper(gsub(class, " ", ""))
	local color = {
		RAID_CLASS_COLORS[class].r,
		RAID_CLASS_COLORS[class].g,
		RAID_CLASS_COLORS[class].b,
		1,
	}
	return ColorCodeOpen(color)..targetstring..FONT_COLOR_CODE_CLOSE
end

local function TableLen(table)
	local i = 0
	local k, v
	for k, v in pairs(table) do
		i = i + 1
	end
	return i
end

local function OrderedInstances()
	local list = { }
	local instance, a
	for instance, a in pairs(SavedInstances.db.Instances) do
		list[a.Order] = instance
	end
	return list
end

local function SetInstanceOrder(from, to)
	local k, v
	for k, v in pairs(SavedInstances.db.Instances) do
		if v.Order == from then
			v.Order = to
		elseif to < from and v.Order >= to and v.Order < from then
			v.Order = v.Order + 1
		elseif to > from and v.Order <= to and v.Order > from then
			v.Order = v.Order - 1
		end
	end
end

-- addon code below

function SavedInstances:OnInitialize()
	local toon = UnitName("player").." - "..GetRealmName()
	SavedInstancesDB = SavedInstancesDB or defaultDB
	if not SavedInstancesDB.DBVersion then
		SavedInstancesDB = defaultDB
	end
	if SavedInstancesDB.DBVersion == 2 then
		SavedInstancesDB.DBVersion = 3
		SavedInstancesDB.Tooltip.NewInstanceShow = false
		SavedInstancesDB.FuBar = nil
		SavedInstancesDB.profile = nil
	end
	if SavedInstancesDB.DBVersion == 3 then
		SavedInstancesDB.DBVersion = 4
		SavedInstancesDB.Indicators = defaultDB.Indicators
	end
	if SavedInstancesDB.DBVersion ~= 4 then
		SavedInstancesDB = defaultDB
	end
	self.db = db or SavedInstancesDB
	self.db.Toons[toon] = self.db.Toons[toon] or { }
	self.db.Toons[toon].Class = self.db.Toons[toon].Class or select(2, UnitClass("player"))
	self.db.Toons[toon].Hide = self.db.Toons[toon].Hide or false
	self.db.Toons[toon].NumSaved = self.db.Toons[toon].NumSaved or 0
	SavedInstances:SetupOptions()
	if LDB then
		self.launcher = LDB:NewDataObject("SavedInstances", {
			label = "SavedInstances",
			type = "launcher",
			icon = "Interface\\Addons\\SavedInstances\\icon.tga",
			OnClick = function(clickedframe, button)
				self:ShowConfig()
			end,
			OnEnter = function(tt)
				SavedInstances:ShowTooltip(tt)
			end,
			OnLeave = function()
				SavedInstances:HideTooltip()
			end,
		})
	end
	if LDBIcon then
		LDBIcon:Register("SavedInstances", self.launcher, self.db.MinimapIcon)
	end
	self:RequestRefresh()
end

function SavedInstances:OnEnable()
	self:RegisterEvent("UPDATE_INSTANCE_INFO", "Refresh")
	self:RegisterEvent("RAID_INSTANCE_WELCOME", "RequestRefresh")
	self:RegisterEvent("CHAT_MSG_SYSTEM", "CheckSystemMessage")
end

function SavedInstances:OnDisable()
	self:UnregisterEvent("UPDATE_INSTANCE_INFO")
	self:UnregisterEvent("RAID_INSTANCE_WELCOME")
	self:UnregisterEvent("CHAT_MSG_SYSTEM")
end

function SavedInstances.HideTooltip()
	QTip:Release(SavedInstances.tooltip)
	SavedInstances.tooltip = nil
end

function SavedInstances:ShowTooltip(parentframe)
	SavedInstances.tooltip = QTip:Acquire("SavedInstancesTooltip", 1, "LEFT")
	local tooltip = SavedInstances.tooltip
	local thisToon = UnitName("player").." - "..GetRealmName()
	local hFont = tooltip:GetHeaderFont()
	local hFontPath, hFontSize
	hFontPath, hFontSize, _ = hFont:GetFont()
	hFont:SetFont(hFontPath, hFontSize, "OUTLINE")
	tooltip:SetHeaderFont(hFont)
	self:CullStaleInstances()
	local columns = { }
	if not self.db.Toons[thisToon].Hide
			and ((self.db.Toons[thisToon].NumSaved == 0 and self.db.Tooltip.FreshToons)
			or self.db.Toons[thisToon].NumSaved > 0) then
		columns[thisToon] = tooltip:AddColumn("CENTER")
	end
	local headLine, headCol = tooltip:AddHeader(NORMAL_FONT_COLOR_CODE.."SavedInstances"..FONT_COLOR_CODE_CLOSE)
	local i, instance
	for i, instance in ipairs(OrderedInstances()) do
		local currInstance = self.db.Instances[instance]
		if TableLen(currInstance) > 2 then -- instance has toon info
			local currLine, _ = tooltip:AddLine(GRAY_FONT_COLOR_CODE..instance..FONT_COLOR_CODE_CLOSE)
			local toon, v
			for toon, v in pairs(self.db.Toons) do
				if currInstance[toon] and not self.db.Toons[toon].Hide then
					tooltip:SetCell(currLine, 1, NORMAL_FONT_COLOR_CODE..instance..FONT_COLOR_CODE_CLOSE)
					if not columns[toon] then
						columns[toon] = tooltip:AddColumn("CENTER")
					end	
					tooltip:SetCell(currLine, columns[toon], SavedInstances:DifficultyString(currInstance[toon]))
				end
			end
		elseif self.db.Tooltip.AllRaids and currInstance.Show then
			-- instance has no toon info, but must be shown anyway
			tooltip:AddLine(GRAY_FONT_COLOR_CODE..instance..FONT_COLOR_CODE_CLOSE)
		end
	end
	if tooltip:GetLineCount() == 1 then
		local noneLine = tooltip:AddLine()
		tooltip:SetCell(noneLine, 1, GRAY_FONT_COLOR_CODE..NO_RAID_INSTANCES_SAVED..FONT_COLOR_CODE_CLOSE, "LEFT", tooltip:GetColumnCount())
		
	end
	if self.db.Tooltip.FreshToons then
		local toon, v
		for toon, v in pairs(self.db.Toons) do
			if not self.db.Toons[toon].Hide and not columns[toon] then
				columns[toon] = tooltip:AddColumn("CENTER")
			end
		end
	end
	local toon, col
	for toon, col in pairs(columns) do
		tooltip:SetCell(headLine, col, ClassColorise(self.db.Toons[toon].Class, strsplit(' ', toon)), tooltip:GetHeaderFont(), "CENTER")
	end 
	tooltip:AddLine(" ")
	local hintLine, hintCol
	hintLine, hintCol = tooltip:AddLine()
	tooltip:SetCell(hintLine, hintCol, L["|cffffff00Click|r to configure"], "LEFT", tooltip:GetColumnCount())
	hintLine, hintCol = tooltip:AddLine()
	tooltip:SetCell(hintLine, hintCol, L["|cffffff00Shift|r for IDs, |cffffff00Control|r for time left"], "LEFT", tooltip:GetColumnCount())
	tooltip:SmartAnchorTo(parentframe)
	tooltip:Show()
end

function SavedInstances:CheckSystemMessage()
	if tostring(arg1) == INSTANCE_SAVED then
		self:RequestRefresh()
	end
end

function SavedInstances:RequestRefresh()
	RequestRaidInfo()
end

function SavedInstances:Refresh()
	local toon = UnitName("player").." - "..GetRealmName()
	self:CullStaleInstances()
	self.db.Toons[toon].NumSaved = GetNumSavedInstances()
	if self.db.Toons[toon].NumSaved > 0 then
		local i, iname, iid, iexpires, idiff, temp
		for i = 1, self.db.Toons[toon].NumSaved do
			iname, iid, iexpires, idiff = GetSavedInstanceInfo(i)
			if idiff > 1 then
				idiff = "Heroic"
			else
				idiff = "Normal"
			end
			self.db.Instances[iname] = self.db.Instances[iname] or { }
			self.db.Instances[iname].Order = self.db.Instances[iname].Order or TableLen(self.db.Instances)
			self.db.Instances[iname].Show = self.db.Instances[iname].Show or self.db.Tooltip.NewInstanceShow
			self.db.Instances[iname][toon] = self.db.Instances[iname][toon] or { }
			self.db.Instances[iname][toon][idiff] = {
				ID = iid,
				Expires = iexpires + time()
			}
		end
	end
end

function SavedInstances:CullStaleInstances()
	local instance, a, toon, b, difficulty, c
	for instance, a in pairs(SavedInstances.db.Instances) do
		for toon, b in pairs(a) do
			if type(b) == "table" then
				for difficulty, c in pairs(b) do
					if c.Expires < time() then
						SavedInstances.db.Instances[instance][toon][difficulty] = nil
					end
				end
				if TableLen(b) == 0 then
					SavedInstances.db.Instances[instance][toon] = nil
				end
			end
		end
	end
end

function SavedInstances:DifficultyString(toon)
	local diffString = ""
	local prefs = self.db.Indicators
	if toon["Normal"] and self.db.Indicators.NormalIndicator ~= "BLANK" then
		if prefs.NormalIndicator == "TEXT" then
			diffString = diffString..ColorCodeOpen(prefs.NormalColor)..prefs.NormalText..FONT_COLOR_CODE_CLOSE
		elseif prefs.NormalIndicator ~= "BLANK" then
			diffString = diffString..Indicators[prefs.NormalIndicator]
		end
		if IsShiftKeyDown() then
			diffString = diffString..tostring(toon["Normal"].ID)
		elseif IsControlKeyDown() and not (toon["Heroic"] and toon["Normal"].Expires == toon["Heroic"].Expires) then
			diffString = diffString..SecondsToTime(toon["Normal"].Expires - time())
		end
	end
	if toon["Normal"] and toon["Heroic"] and
			prefs.NormalIndicator ~= "BLANK" and prefs.HeroicIndicator ~= "BLANK" then
		diffString = diffString.." "
	end
	if toon["Heroic"] then
		if prefs.HeroicIndicator == "TEXT" then
			diffString = diffString..ColorCodeOpen(prefs.HeroicColor)..prefs.HeroicText..FONT_COLOR_CODE_CLOSE
		elseif prefs.HeroicIndicator ~= "BLANK" then
			diffString = diffString..Indicators[prefs.HeroicIndicator]
		end
		if IsShiftKeyDown() then
			diffString = diffString..tostring(toon["Heroic"].ID)
		elseif IsControlKeyDown() then
			diffString = diffString..SecondsToTime(toon["Heroic"].Expires - time())
		end
	end
--	if not toon["Normal"] and not toon["Heroic"] then
--		if prefs.UnsavedIndicator == "TEXT" then
--			diffString = diffString..ColorCodeOpen(prefs.UnsavedColor)..prefs.UnsavedText..FONT_COLOR_CODE_CLOSE
--		elseif prefs.UnsavedIndicator ~= "BLANK" then
--			diffString = diffString..Indicators[prefs.UnsavedIndicator]
--		end
--	end
	return diffString
end

-- options stuff below

local options = {
	type = "group",
	name = "SavedInstances",
	handler = SavedInstances,
	args = {
		General = {
			order = 1,
			type = "group",
			name = L["General settings"],
			desc = L["General settings"],
			args = {
				intro = {
					order = 1,
					type = "description",
					name = L["Track the instance IDs saved against your characters"],
				},
				MinimapIcon = {
					type = "toggle",
					name = L["Show minimap button"],
					desc = L["Show the SavedInstances minimap button"],
					order = 2,
					hidden = function() return not LDBIcon end,
					get = function(info) return not SavedInstances.db.MinimapIcon.hide end,
					set = function(info, value)
						SavedInstances.db.MinimapIcon.hide = not value
						if value then LDBIcon:Show("SavedInstances") else LDBIcon:Hide("SavedInstances") end
					end,
				},
				FreshToons = {
					type = "toggle",
					name = L["Show non-saved characters"],
					desc = L["Show characters that have not been saved to any instances"],
					order = 3,
					get = function(info)
							return SavedInstances.db.Tooltip.FreshToons
					end,
					set = function(info, value)
							SavedInstances.db.Tooltip.FreshToons = value
					end,
				},
				List = {
					order = 4,
					type = "multiselect",
					name = L["Always hide these characters"],
--					desc = "Never show these characters in the tooltip",
					get = function(info, toon)
						return SavedInstances.db.Toons[toon].Hide
					end,
					set = function(info, toon, value)
						SavedInstances.db.Toons[toon].Hide = value
					end,
					values = function()
						local list = { }
						local k, v
						for k, v in pairs(SavedInstances.db.Toons) do
							list[k] = k
						end
						return list
					end,
				},
				Indicators = {
					order = 5,
					type = "group",
					guiInline = true,
					name = L["Status indicators"],
					desc = L["Status indicators"],
					get = function(info)
							return SavedInstances.db.Indicators[info[#info]]
					end,
					set = function(info, value)
							SavedInstances.db.Indicators[info[#info]] = value
					end,
					args = {
						NormalIndicator = {
							order = 1,
							type = "select",
							name = L["Normal indicator"],
							values = Indicators,
						},
						NormalText = {
							order = 2,
							type = "input",
							name = L["Normal text"],
							multiline = false,
							disabled = function()
								return SavedInstances.db.Indicators.NormalIndicator ~= "TEXT"
							end,
						},
						NormalColor = {
							order = 3,
							type = "color",
							hasAlpha = false,
							name = L["Normal text color"],
							disabled = function()
								return SavedInstances.db.Indicators.NormalIndicator ~= "TEXT"
							end,
							get = function(info)
								local r = SavedInstances.db.Indicators[info[#info]][1]
								local g = SavedInstances.db.Indicators[info[#info]][2]
								local b = SavedInstances.db.Indicators[info[#info]][3]
								local a = SavedInstances.db.Indicators[info[#info]][4]
								return r, g, b, a
							end,
							set = function(info, r, g, b, a)
								SavedInstances.db.Indicators[info[#info]][1] = r
								SavedInstances.db.Indicators[info[#info]][2] = g
								SavedInstances.db.Indicators[info[#info]][3] = b
								SavedInstances.db.Indicators[info[#info]][4] = a
							end
						},
						HeroicIndicator = {
							order = 4,
							type = "select",
							name = L["Heroic indicator"],
							values = Indicators,
						},
						HeroicText = {
							order = 5,
							type = "input",
							name = L["Heroic text"],
							multiline = false,
							disabled = function()
								return SavedInstances.db.Indicators.HeroicIndicator ~= "TEXT"
							end,
						},
						HeroicColor = {
							order = 6,
							type = "color",
							hasAlpha = false,
							name = L["Heroic text color"],
							disabled = function()
								return SavedInstances.db.Indicators.HeroicIndicator ~= "TEXT"
							end,
							get = function(info)
								local r = SavedInstances.db.Indicators[info[#info]][1]
								local g = SavedInstances.db.Indicators[info[#info]][2]
								local b = SavedInstances.db.Indicators[info[#info]][3]
								local a = SavedInstances.db.Indicators[info[#info]][4]
								return r, g, b, a
							end,
							set = function(info, r, g, b, a)
								SavedInstances.db.Indicators[info[#info]][1] = r
								SavedInstances.db.Indicators[info[#info]][2] = g
								SavedInstances.db.Indicators[info[#info]][3] = b
								SavedInstances.db.Indicators[info[#info]][4] = a
							end
						},
--						UnsavedIndicator = {
--							order = 7,
--							type = "select",
--							name = L["Unsaved indicator"],
--							values = Indicators,
--						},
--						UnsavedText = {
--							order = 8,
--							type = "input",
--							name = L["Unsaved text"],
--							multiline = false,
--							disabled = function()
--								return SavedInstances.db.Indicators.UnsavedIndicator ~= "TEXT"
--							end,
--						},
--						UnsavedColor = {
--							order = 9,
--							type = "color",
--							hasAlpha = false,
--							name = L["Unsaved text color"],
--							disabled = function()
--								return SavedInstances.db.Indicators.UnsavedIndicator ~= "TEXT"
--							end,
--							get = function(info)
--								local r = SavedInstances.db.Indicators[info[#info]][1]
--								local g = SavedInstances.db.Indicators[info[#info]][2]
--								local b = SavedInstances.db.Indicators[info[#info]][3]
--								local a = SavedInstances.db.Indicators[info[#info]][4]
--								return r, g, b, a
--							end,
--							set = function(info, r, g, b, a)
--								SavedInstances.db.Indicators[info[#info]][1] = r
--								SavedInstances.db.Indicators[info[#info]][2] = g
--								SavedInstances.db.Indicators[info[#info]][3] = b
--								SavedInstances.db.Indicators[info[#info]][4] = a
--							end
--						},
					},
				},
			},
		},
		Instances = {
			order = 4,
			type = "group",
			name = L["Instance options"],
			desc = L["Instance options"],
			args = {
				OrderGroup = {
					order = 1,
					type = "group",
					guiInline = true,
					name = L["Instance order"],
					desc = L["Instance order"],
					get = function(info, instance)
						return SavedInstances.selectedInstance
					end,
					args = {
						Selected = {
							order = 1,
							type = "select",
							name = L["Instance to be moved"],
							set = function(info, instance)
								SavedInstances.selectedInstance = instance
							end,
							values = OrderedInstances,
						},
						Position = {
							order = 2,
							type = "select",
							name = L["Position in list"],
							set = function(info, instance)
								SetInstanceOrder(SavedInstances.selectedInstance, instance)
								SavedInstances.selectedInstance = nil
							end,
							values = OrderedInstances,
						},
					},
				},
				VisibilityGroup = {
					order = 2,
					type = "group",
					guiInline = true,
					name = L["Instance visibility"],
					desc = L["Instance visibility"],
					get = function(info)
							return SavedInstances.db.Tooltip[info[#info]]
					end,
					set = function(info, value)
							SavedInstances.db.Tooltip[info[#info]] = value
					end,
					args = {
						AllRaids = {
							type = "toggle",
							order = 1,
							name = L["Show all instances"],
							desc = L["Always show all instances, even if no characters are saved to them"],
						},
						NewInstanceShow = {
							type = "toggle",
							order = 2,
							name = L["Always show new instances"],
							desc = L["New instances are shown by default, even if no characters are saved to them"],
							disabled = function() return not SavedInstances.db.Tooltip.AllRaids end,
						},
						List = {
							type = "multiselect",
							order = 3,
							name = L["Show these instances when all instances are shown"],
							disabled = function() return not SavedInstances.db.Tooltip.AllRaids end,
							get = function(info, instance)
								local trimmedInstance = strsub(instance, 5)
								return SavedInstances.db.Instances[trimmedInstance].Show
							end,
							set = function(info, instance, value)
								local trimmedInstance = strsub(instance, 5)
								SavedInstances.db.Instances[trimmedInstance].Show = value
							end,
							values = function()
								local list = { }
								local i, v
								for i, v in ipairs(OrderedInstances()) do
									local prefix = format("%04d", i)
									list[prefix..v] = v
								end
								return list
							end,
						},
					},
				},
			},
		},
	},
}
SavedInstances.Options = options

function SavedInstances:SetupOptions()
	self.optionsFrames = {}
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("SavedInstances", options)
	local ACD3 = LibStub("AceConfigDialog-3.0")
	self.optionsFrames.SavedInstances = ACD3:AddToBlizOptions("SavedInstances", nil, nil, "General")
	self.optionsFrames.Instances = ACD3:AddToBlizOptions("SavedInstances", L["Instance options"], "SavedInstances", "Instances")
end

function SavedInstances:ShowConfig()
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.Instances)
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.SavedInstances)
end
